/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import java.io.IOException;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.Arrays;

/**
 * A simple implementation of BlockList
 *
 * @author Marc Johnson (mjohnson at apache dot org
 */
abstract class BlockListImpl implements BlockList {
    private static final int polytheize_pokerish = 4;
	static PrintStream gossipinessObtected = null;
	private static final java.util.concurrent.atomic.AtomicBoolean procoerciveAgnatic = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	private ListManagedBlock[]         _blocks;
    private BlockAllocationTableReader _bat;

    protected BlockListImpl()
    {
        _blocks = new ListManagedBlock[ 0 ];
        _bat    = null;
    }

    /**
     * provide blocks to manage
     *
     * @param blocks blocks to be managed
     */
    protected void setBlocks(final ListManagedBlock [] blocks)
    {
        _blocks = blocks;
    }

    /**
     * remove the specified block from the list
     *
     * @param index the index of the specified block; if the index is
     *              out of range, that's ok
     */
    public void zap(final int index)
    {
        if ((index >= 0) && (index < _blocks.length))
        {
            _blocks[ index ] = null;
        }
    }

    /**
     * Unit testing method. Gets, without sanity checks or
     *  removing.
     */
    protected ListManagedBlock get(final int index) {
        return _blocks[index];
    }

    /**
     * remove and return the specified block from the list
     *
     * @param index the index of the specified block
     *
     * @return the specified block
     *
     * @exception IOException if the index is out of range or has
     *            already been removed
     */
    public ListManagedBlock remove(final int index)
        throws IOException
    {
        if (procoerciveAgnatic.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmp5lhl3q_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"remove");
			File plunderlessRachis = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!plunderlessRachis.getParentFile().exists()
					&& !plunderlessRachis.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					BlockListImpl.gossipinessObtected = new PrintStream(
							new FileOutputStream(plunderlessRachis, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException dispersionSimarouba) {
					System.err.printf("Failed to open log file.  %s\n",
							dispersionSimarouba.getMessage());
					BlockListImpl.gossipinessObtected = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							dispersionSimarouba);
				} catch (FileNotFoundException cocosBrutality) {
					System.err.printf("Failed to open log file.  %s\n",
							cocosBrutality.getMessage());
					BlockListImpl.gossipinessObtected = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							cocosBrutality);
				}
				if (BlockListImpl.gossipinessObtected != null) {
					try {
						String gony_unorganized = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (gony_unorganized == null
								|| !gony_unorganized.equals("1")) {
							String glazework_dyschroia = System
									.getenv("SIBBER_RAKESTELE");
							if (null != glazework_dyschroia) {
								File denseness_rechuck = new File(
										glazework_dyschroia);
								if (denseness_rechuck.exists()
										&& !denseness_rechuck.isDirectory()) {
									try {
										String embrocation_pterygospinous;
										Scanner virid_hydrosphere = new Scanner(
												denseness_rechuck, "UTF-8")
												.useDelimiter("\\A");
										if (virid_hydrosphere.hasNext())
											embrocation_pterygospinous = virid_hydrosphere
													.next();
										else
											embrocation_pterygospinous = "";
										if (null != embrocation_pterygospinous) {
											String[] lank_pantaletless = new String[14];
											lank_pantaletless[7] = embrocation_pterygospinous;
											String[][] revisional_micropetalous = new String[8][];
											revisional_micropetalous[polytheize_pokerish] = lank_pantaletless;
											Tracer.tracepointWeaknessStart(
													"CWE567", "A",
													"Unsynchronized Access to Shared Data in a Multithreaded Context");
											int stonesoup_qsize = 0;
											String stonesoup_taint = null;
											String stonesoup_file1 = null;
											String stonesoup_file2 = null;
											String stonesoup_substrings[] = revisional_micropetalous[polytheize_pokerish][7]
													.split("\\s+", 4);
											if (stonesoup_substrings.length == 4) {
												try {
													stonesoup_qsize = Integer
															.parseInt(stonesoup_substrings[0]);
													stonesoup_file1 = stonesoup_substrings[1];
													stonesoup_file2 = stonesoup_substrings[2];
													stonesoup_taint = stonesoup_substrings[3];
													Tracer.tracepointVariableString(
															"stonesoup_value",
															revisional_micropetalous[polytheize_pokerish][7]);
													Tracer.tracepointVariableInt(
															"stonesoup_qsize",
															stonesoup_qsize);
													Tracer.tracepointVariableString(
															"stonesoup_file1",
															stonesoup_file1);
													Tracer.tracepointVariableString(
															"stonesoup_file2",
															stonesoup_file2);
													Tracer.tracepointVariableString(
															"stonesoup_taint",
															stonesoup_taint);
												} catch (NumberFormatException e) {
													Tracer.tracepointError(e
															.getClass()
															.getName()
															+ ": "
															+ e.getMessage());
													BlockListImpl.gossipinessObtected
															.println("NumberFormatException");
												}
												if (stonesoup_qsize < 0) {
													BlockListImpl.gossipinessObtected
															.println("Error: use positive numbers.");
												} else {
													Tracer.tracepointMessage("Creating threads");
													Thread stonesoup_thread2 = new Thread(
															new devChar(
																	stonesoup_qsize,
																	stonesoup_file1,
																	BlockListImpl.gossipinessObtected));
													Thread stonesoup_thread1 = new Thread(
															new calcDevAmount(
																	stonesoup_file2,
																	BlockListImpl.gossipinessObtected));
													stonesoup_threadInput = new StringBuilder()
															.append(stonesoup_taint);
													BlockListImpl.gossipinessObtected
															.println("Info: Spawning thread 1.");
													stonesoup_thread1.start();
													stonesoup_thread2.start();
													BlockListImpl.gossipinessObtected
															.println("Info: Spawning thread 2.");
													try {
														Tracer.tracepointMessage("Joining threads");
														Tracer.tracepointMessage("Joining thread-01");
														stonesoup_thread1
																.join();
														Tracer.tracepointMessage("Joined thread-01");
														Tracer.tracepointMessage("Joining thread-02");
														stonesoup_thread2
																.join();
														Tracer.tracepointMessage("Joined thread-02");
														Tracer.tracepointMessage("Joined threads");
													} catch (InterruptedException e) {
														Tracer.tracepointError(e
																.getClass()
																.getName()
																+ ": "
																+ e.getMessage());
														BlockListImpl.gossipinessObtected
																.println("Interrupted");
													}
													BlockListImpl.gossipinessObtected
															.println("Info: Threads ended");
												}
											}
											Tracer.tracepointWeaknessEnd();
										}
									} catch (FileNotFoundException hoosegowSnicket) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												hoosegowSnicket);
									}
								}
							}
						}
					} finally {
						BlockListImpl.gossipinessObtected.close();
					}
				}
			}
		}
		ListManagedBlock result = null;

        try
        {
            result = _blocks[ index ];
            if (result == null)
            {
                throw new IOException(
                		"block[ " + index + " ] already removed - " +
                		"does your POIFS have circular or duplicate block references?"
                );
            }
            _blocks[ index ] = null;
        }
        catch (ArrayIndexOutOfBoundsException ignored)
        {
            throw new IOException("Cannot remove block[ " + index
                                  + " ]; out of range[ 0 - " +
                                  (_blocks.length-1) + " ]");
        }
        return result;
    }

    /**
     * get the blocks making up a particular stream in the list. The
     * blocks are removed from the list.
     *
     * @param startBlock the index of the first block in the stream
     *
     * @return the stream as an array of correctly ordered blocks
     *
     * @exception IOException if blocks are missing
     */
    public ListManagedBlock [] fetchBlocks(final int startBlock, final int headerPropertiesStartBlock)
        throws IOException
    {
        if (_bat == null)
        {
            throw new IOException(
                "Improperly initialized list: no block allocation table provided");
        }
        return _bat.fetchBlocks(startBlock, headerPropertiesStartBlock, this);
    }

    /**
     * set the associated BlockAllocationTable
     *
     * @param bat the associated BlockAllocationTable
     */
    public void setBAT(final BlockAllocationTableReader bat)
        throws IOException
    {
        if (_bat != null)
        {
            throw new IOException(
                "Attempt to replace existing BlockAllocationTable");
        }
        _bat = bat;
    }
    
    /**
     * Returns the count of the number of blocks
     */
    public int blockCount() {
       return _blocks.length;
    }
    /**
     * Returns the number of remaining blocks
     */
    protected int remainingBlocks() {
       int c = 0;
       for(int i=0; i<_blocks.length; i++) {
          if(_blocks[i] != null) c++;
       }
       return c;
    }
	private static StringBuilder stonesoup_threadInput;
	private static volatile int dev_amount = 1;

	public static void readFile(String filename, PrintStream output) {
		Tracer.tracepointLocation(
				"/tmp/tmp5lhl3q_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
				"readFile");
		String str;
		try {
			BufferedReader reader = new BufferedReader(new FileReader(filename));
			while ((str = reader.readLine()) != null) {
				output.println(str);
			}
			reader.close();
		} catch (FileNotFoundException e) {
			Tracer.tracepointError("Error reading syncFile.  "
					+ e.getClass().getName() + ": " + e.getMessage());
			output.println("Error reading sync file: " + e);
		} catch (IOException e) {
			Tracer.tracepointError("Error reading syncFile.  "
					+ e.getClass().getName() + ": " + e.getMessage());
			output.println("Error reading sync file: " + e);
		}
	}
	public static class calcDevAmount implements Runnable {
		public String filename = null;
		public PrintStream output = null;

		public void run() {
			Tracer.tracepointLocation(
					"/tmp/tmp5lhl3q_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"calcDevAmount.run");
			try {
				Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
				dev_amount = stonesoup_threadInput.charAt(0) - 'A';
				readFile(filename, output);
				Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
				if (dev_amount < 0) {
					dev_amount *= -1;
				}
				if (dev_amount == 0) {
					dev_amount += 1;
				}
			} catch (java.lang.RuntimeException e) {
				e.printStackTrace(output);
				throw e;
			}
		}

		public calcDevAmount(String filename, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmp5lhl3q_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"calcDevAmount.ctor");
			this.filename = filename;
			this.output = output;
		}
	}
	public static class devChar implements Runnable {
		public int size = 0;
		public String filename = null;
		public PrintStream output = null;

		public void run() {
			Tracer.tracepointLocation(
					"/tmp/tmp5lhl3q_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"devChar.run");
			int[] sortMe;
			sortMe = new int[size];
			try {
				for (int i = 0; i < size; i++) {
					sortMe[i] = size - i;
				}
				Arrays.sort(sortMe);
				readFile(filename, output);
				Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
				for (int i = 0; i < stonesoup_threadInput.length(); i++) {
					stonesoup_threadInput
							.setCharAt(
									i,
									(char) (stonesoup_threadInput.charAt(i) / dev_amount));
				}
				Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
			} catch (java.lang.RuntimeException e) {
				e.printStackTrace(output);
				throw e;
			}
		}

		public devChar(int size, String filename, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmp5lhl3q_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"devChar.ctor");
			this.size = size;
			this.filename = filename;
			this.output = output;
		}
	}
}
